package cz.fidentis.analyst.visitors.mesh;

import cz.fidentis.analyst.mesh.core.CornerTableRow;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshFacetImpl;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import cz.fidentis.analyst.mesh.core.MeshPointImpl;
import cz.fidentis.analyst.mesh.core.MeshTriangle;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;
import javax.vecmath.Vector3d;
import org.junit.jupiter.api.Assertions;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import org.junit.jupiter.api.Test;

/**
 *
 * @author oslejsek
 */
public class Point2MeshTriVisitorTest {
    private MeshFacet getTrivialFacet(double offset, double size) {
        MeshFacet facet = new MeshFacetImpl();
        facet.addVertex(new MeshPointImpl(new Vector3d(0, 0, offset), new Vector3d(0, 0, 1), new Vector3d()));
        facet.addVertex(new MeshPointImpl(new Vector3d(size, 0, offset), new Vector3d(0, 0, 1), new Vector3d()));
        facet.addVertex(new MeshPointImpl(new Vector3d(0, size, offset), new Vector3d(0, 0, 1), new Vector3d()));

        facet.getCornerTable().addRow(new CornerTableRow(0, -1));
        facet.getCornerTable().addRow(new CornerTableRow(1, -1));
        facet.getCornerTable().addRow(new CornerTableRow(2, -1));

        return facet;
    }

    @Test
    public void absoluteDistTest() {
        MeshFacet facet = getTrivialFacet(1.5, 1);
        MeshPoint point = new MeshPointImpl(new Vector3d(0,0,0), new Vector3d(0,0,-1), new Vector3d());
        
        Point2MeshVisitor vis = new Point2MeshTriVisitor(point, false);
        facet.accept(vis);
        assertEquals(1.5, vis.getDistance());
        
        List<Integer> closestPoints = vis.getIndices();
        List<MeshFacet> closestMeshes = vis.getClosestFacets();
        assertNotNull(closestPoints);
        assertNotNull(closestMeshes);
        assertEquals(1, closestPoints.size());
        assertEquals(1, closestMeshes.size());
        assertEquals(facet, closestMeshes.get(0));
        assertEquals(0, closestPoints.get(0));
    }
    
    @Test
    public void relativeDistTest() {
        MeshFacet facet = getTrivialFacet(1.5, 1);
        MeshPoint point = new MeshPointImpl(new Vector3d(0,0,0), new Vector3d(0,0,-1), new Vector3d());
        
        Point2MeshVisitor vis = new Point2MeshTriVisitor(point, true);
        facet.accept(vis);
        assertEquals(-1.5, vis.getDistance());
        
        List<Integer> closestPoints = vis.getIndices();
        List<MeshFacet> closestMeshes = vis.getClosestFacets();
        assertNotNull(closestPoints);
        assertNotNull(closestMeshes);
        assertEquals(1, closestPoints.size());
        assertEquals(1, closestMeshes.size());
        assertEquals(facet, closestMeshes.get(0));
        assertEquals(0, closestPoints.get(0));
    }
    
    @Test
    public void exactMatchTest() {
        MeshFacet facet = getTrivialFacet(1.5, 1);
        MeshPoint point = new MeshPointImpl(new Vector3d(0, 0, 1.5), new Vector3d(0, 0, 1), new Vector3d());
        
        System.out.println(facet.getVertices());
        
        Point2MeshVisitor vis = new Point2MeshTriVisitor(point, false);
        facet.accept(vis);
        assertEquals(0, vis.getDistance());
        
        System.out.println(facet.getVertices());
        
        List<Integer> closestPoints = vis.getIndices();
        List<MeshFacet> closestMeshes = vis.getClosestFacets();
        assertNotNull(closestPoints);
        assertNotNull(closestMeshes);
        assertEquals(1, closestPoints.size());
        assertEquals(1, closestMeshes.size());
        assertEquals(facet, closestMeshes.get(0));
        assertEquals(0, closestPoints.get(0));
    }
    
    @Test
    public void getProjectionToTrianglePlaneTest()
            throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        Class<?>[] argClasses = new Class[2];
        argClasses[0] = Vector3d.class;
        argClasses[1] = MeshTriangle.class;
        Method method = Point2MeshTriVisitor.class.getDeclaredMethod("getProjectionToTrianglePlane", argClasses);
        method.setAccessible(true);
        Object[] args = new Object[2];

        args[0] = new Vector3d(0, 0.5, 1);
        MeshTriangle triangle = new MeshTriangle(
                new MeshPointImpl(new Vector3d(1, 0, 0), null, null),
                new MeshPointImpl(new Vector3d(1, 2, 0), null, null),
                new MeshPointImpl(new Vector3d(2, 0, 0), null, null)
        );
        args[1] = triangle;

        Point2MeshTriVisitor hausdorffDistance = new Point2MeshTriVisitor(new MeshPointImpl(new Vector3d(), null, null), false);
        Vector3d result = (Vector3d) method.invoke(hausdorffDistance, args);

        Assertions.assertEquals(0d, result.x);
        Assertions.assertEquals(0.5d, result.y);
        Assertions.assertEquals(0d, result.z);
    }

    @Test
    public void getProjectionToTrianglePlaneBelowPLaneTest()
            throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        Class<?>[] argClasses = new Class[2];
        argClasses[0] = Vector3d.class;
        argClasses[1] = MeshTriangle.class;
        Method method = Point2MeshTriVisitor.class.getDeclaredMethod("getProjectionToTrianglePlane", argClasses);
        method.setAccessible(true);
        Object[] args = new Object[2];

        args[0] = new Vector3d(0, 0.5, 1);
        MeshTriangle triangle = new MeshTriangle(
                new MeshPointImpl(new Vector3d(1, 0, 0), null, null),
                new MeshPointImpl(new Vector3d(1, 2, 0), null, null),
                new MeshPointImpl(new Vector3d(2, 0, 0), null, null)
        );
        args[1] = triangle;

        Point2MeshTriVisitor hausdorffDistance = new Point2MeshTriVisitor(new MeshPointImpl(new Vector3d(), null, null), false);
        Vector3d result = (Vector3d) method.invoke(hausdorffDistance, args);

        Assertions.assertEquals(0d, result.x);
        Assertions.assertEquals(0.5d, result.y);
        Assertions.assertEquals(0d, result.z);
    }

    @Test
    public void isPointInTriangleInsideTriangleTest()
            throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        Class<?>[] argClasses = new Class[2];
        argClasses[0] = Vector3d.class;
        argClasses[1] = MeshTriangle.class;
        Method method = Point2MeshTriVisitor.class.getDeclaredMethod("isPointInTriangle", argClasses);
        method.setAccessible(true);
        Object[] args = new Object[2];

        args[0] = new Vector3d(1.2, 0.2, 0);
        MeshTriangle triangle = new MeshTriangle(
                new MeshPointImpl(new Vector3d(1, 0, 0), null, null),
                new MeshPointImpl(new Vector3d(1, 2, 0), null, null),
                new MeshPointImpl(new Vector3d(2, 0, 0), null, null)
        );
        args[1] = triangle;

        Point2MeshTriVisitor hausdorffDistance = new Point2MeshTriVisitor(new MeshPointImpl(new Vector3d(), null, null), false);
        boolean result = (boolean) method.invoke(hausdorffDistance, args);

        Assertions.assertTrue(result);
    }

    @Test
    public void isPointInTriangleOutsideTriangleTest()
            throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        Class<?>[] argClasses = new Class[2];
        argClasses[0] = Vector3d.class;
        argClasses[1] = MeshTriangle.class;
        Method method = Point2MeshTriVisitor.class.getDeclaredMethod("isPointInTriangle", argClasses);
        method.setAccessible(true);
        Object[] args = new Object[2];

        args[0] = new Vector3d(0, 0.5, 0);
        MeshTriangle triangle = new MeshTriangle(
                new MeshPointImpl(new Vector3d(1, 0, 0), null, null),
                new MeshPointImpl(new Vector3d(1, 2, 0), null, null),
                new MeshPointImpl(new Vector3d(2, 0, 0), null, null)
        );
        args[1] = triangle;

        Point2MeshTriVisitor hausdorffDistance = new Point2MeshTriVisitor(new MeshPointImpl(new Vector3d(), null, null), false);
        boolean result = (boolean) method.invoke(hausdorffDistance, args);

        Assertions.assertFalse(result);
    }

    @Test
    public void getProjectionToClosestEdgeTest()
            throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        Class<?>[] argClasses = new Class[2];
        argClasses[0] = Vector3d.class;
        argClasses[1] = MeshTriangle.class;
        Method method = Point2MeshTriVisitor.class.getDeclaredMethod("getProjectionToClosestEdge", argClasses);
        method.setAccessible(true);
        Object[] args = new Object[2];

        args[0] = new Vector3d(0, 0.5, 0);
        MeshTriangle triangle = new MeshTriangle(
                new MeshPointImpl(new Vector3d(1, 0, 0), null, null),
                new MeshPointImpl(new Vector3d(1, 2, 0), null, null),
                new MeshPointImpl(new Vector3d(2, 0, 0), null, null)
        );
        args[1] = triangle;

        Point2MeshTriVisitor hausdorffDistance = new Point2MeshTriVisitor(new MeshPointImpl(new Vector3d(), null, null), false);
        Vector3d result = (Vector3d) method.invoke(hausdorffDistance, args);

        Assertions.assertEquals(1d, result.x);
        Assertions.assertEquals(0.5d, result.y);
        Assertions.assertEquals(0d, result.z);
    }

    @Test
    public void getProjectionToClosestEdgeNormalProjectionOutsideEdgeTestTest()
            throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        Class<?>[] argClasses = new Class[2];
        argClasses[0] = Vector3d.class;
        argClasses[1] = MeshTriangle.class;
        Method method = Point2MeshTriVisitor.class.getDeclaredMethod("getProjectionToClosestEdge", argClasses);
        method.setAccessible(true);
        Object[] args = new Object[2];

        args[0] = new Vector3d(0, -0.5, 0);
        MeshTriangle triangle = new MeshTriangle(
                new MeshPointImpl(new Vector3d(1, 0, 0), null, null),
                new MeshPointImpl(new Vector3d(1, 2, 0), null, null),
                new MeshPointImpl(new Vector3d(2, 0, 0), null, null)
        );
        args[1] = triangle;

        Point2MeshTriVisitor hausdorffDistance = new Point2MeshTriVisitor(new MeshPointImpl(new Vector3d(), null, null), false);
        Vector3d result = (Vector3d) method.invoke(hausdorffDistance, args);

        Assertions.assertEquals(1d, result.x);
        Assertions.assertEquals(0d, result.y);
        Assertions.assertEquals(0d, result.z);
    }

    @Test
    public void getProjectionToEdgeTest()
            throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        Class<?>[] argClasses = new Class[3];
        argClasses[0] = Vector3d.class;
        argClasses[1] = Vector3d.class;
        argClasses[2] = Vector3d.class;
        Method method = Point2MeshTriVisitor.class.getDeclaredMethod("getProjectionToEdge", argClasses);
        method.setAccessible(true);
        Object[] args = new Object[3];

        args[0] = new Vector3d(0, 0.5, 0);
        args[1] = new Vector3d(1, 2, 0);
        args[2] = new Vector3d(1, 0, 0);

        Point2MeshTriVisitor hausdorffDistance = new Point2MeshTriVisitor(new MeshPointImpl(new Vector3d(), null, null), false);
        Vector3d result = (Vector3d) method.invoke(hausdorffDistance, args);

        Assertions.assertEquals(1d, result.x);
        Assertions.assertEquals(0.5d, result.y);
        Assertions.assertEquals(0d, result.z);
    }

    @Test
    public void getProjectionToEdgeNormalProjectionOutsideEdgeTest()
            throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        Class<?>[] argClasses = new Class[3];
        argClasses[0] = Vector3d.class;
        argClasses[1] = Vector3d.class;
        argClasses[2] = Vector3d.class;
        Method method = Point2MeshTriVisitor.class.getDeclaredMethod("getProjectionToEdge", argClasses);
        method.setAccessible(true);
        Object[] args = new Object[3];

        args[0] = new Vector3d(0, -0.5, 0);
        args[1] = new Vector3d(1, 2, 0);
        args[2] = new Vector3d(1, 0, 0);

        Point2MeshTriVisitor hausdorffDistance = new Point2MeshTriVisitor(new MeshPointImpl(new Vector3d(), null, null), false);
        Vector3d result = (Vector3d) method.invoke(hausdorffDistance, args);

        Assertions.assertEquals(1d, result.x);
        Assertions.assertEquals(0d, result.y);
        Assertions.assertEquals(0d, result.z);
    }

    @Test
    public void getProjectionToEdgeNormalProjectionOutsideEdge2Test()
            throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        Class<?>[] argClasses = new Class[3];
        argClasses[0] = Vector3d.class;
        argClasses[1] = Vector3d.class;
        argClasses[2] = Vector3d.class;
        Method method = Point2MeshTriVisitor.class.getDeclaredMethod("getProjectionToEdge", argClasses);
        method.setAccessible(true);
        Object[] args = new Object[3];

        args[0] = new Vector3d(0, 2.5, 0);
        args[1] = new Vector3d(1, 2, 0);
        args[2] = new Vector3d(1, 0, 0);

        Point2MeshTriVisitor hausdorffDistance = new Point2MeshTriVisitor(new MeshPointImpl(new Vector3d(), null, null), false);
        Vector3d result = (Vector3d) method.invoke(hausdorffDistance, args);

        Assertions.assertEquals(1d, result.x);
        Assertions.assertEquals(2d, result.y);
        Assertions.assertEquals(0d, result.z);
    }

}
